# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.arangodb.operators.arangodb import AQLOperator
from airflow.providers.arangodb.sensors.arangodb import AQLSensor

dag = DAG(
    "example_arangodb_operator",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
)

# [START howto_aql_sensor_arangodb]

sensor = AQLSensor(
    task_id="aql_sensor",
    query="FOR doc IN students FILTER doc.name == 'judy' RETURN doc",
    timeout=60,
    poke_interval=10,
    dag=dag,
)

# [END howto_aql_sensor_arangodb]

# [START howto_aql_sensor_template_file_arangodb]

sensor2 = AQLSensor(
    task_id="aql_sensor_template_file",
    query="search_judy.sql",
    timeout=60,
    poke_interval=10,
    dag=dag,
)

# [END howto_aql_sensor_template_file_arangodb]


# [START howto_aql_operator_arangodb]

operator = AQLOperator(
    task_id="aql_operator",
    query="FOR doc IN students RETURN doc",
    dag=dag,
    result_processor=lambda cursor: print([document["name"] for document in cursor]),
)

# [END howto_aql_operator_arangodb]

# [START howto_aql_operator_template_file_arangodb]

operator2 = AQLOperator(
    task_id="aql_operator_template_file",
    dag=dag,
    result_processor=lambda cursor: print([document["name"] for document in cursor]),
    query="search_all.sql",
)

# [END howto_aql_operator_template_file_arangodb]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any

from airflow.exceptions import AirflowException
from airflow.providers.arangodb.hooks.arangodb import ArangoDBHook
from airflow.providers.arangodb.version_compat import BaseOperator

if TYPE_CHECKING:
    try:
        from airflow.sdk.definitions.context import Context
    except ImportError:
        # TODO: Remove once provider drops support for Airflow 2
        from airflow.utils.context import Context


class AQLOperator(BaseOperator):
    """
    Executes AQL query in a ArangoDB database.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AQLOperator`

    :param query: the AQL query to be executed. Can receive a str representing a
        AQL statement, or you can provide .sql file having the query
    :param result_processor: function to further process the Result from ArangoDB
    :param arangodb_conn_id: Reference to :ref:`ArangoDB connection id <howto/connection:arangodb>`.
    """

    template_fields: Sequence[str] = ("query",)
    template_ext: Sequence[str] = (".sql",)
    template_fields_renderers = {"query": "sql"}

    def __init__(
        self,
        *,
        query: str,
        arangodb_conn_id: str = "arangodb_default",
        result_processor: Callable | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.arangodb_conn_id = arangodb_conn_id
        self.query = query
        self.result_processor = result_processor

    def execute(self, context: Context):
        self.log.info("Executing: %s", self.query)
        hook = ArangoDBHook(arangodb_conn_id=self.arangodb_conn_id)
        result = hook.query(self.query)
        if self.result_processor:
            self.result_processor(result)


class ArangoDBCollectionOperator(BaseOperator):
    """
    Executes collection operations in a ArangoDB database.

    :param arangodb_conn_id: Connection ID for ArangoDB, defaults to "arangodb_default".
    :param collection_name: The name of the collection to be operated on.
    :param documents_to_insert: A list of python dictionaries to insert into the collection.
    :param documents_to_update: A list of python dictionaries to update in the collection.
    :param documents_to_replace: A list of python dictionaries to replace in the collection.
    :param documents_to_delete: A list of python dictionaries to delete from the collection.
    :param delete_collection: If True, the specified collection will be deleted.
    """

    def __init__(
        self,
        *,
        arangodb_conn_id: str = "arangodb_default",
        collection_name: str,
        documents_to_insert: list[dict[str, Any]] | None = None,
        documents_to_update: list[dict[str, Any]] | None = None,
        documents_to_replace: list[dict[str, Any]] | None = None,
        documents_to_delete: list[dict[str, Any]] | None = None,
        delete_collection: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.arangodb_conn_id = arangodb_conn_id
        self.collection_name = collection_name
        self.documents_to_insert = documents_to_insert or []
        self.documents_to_update = documents_to_update or []
        self.documents_to_replace = documents_to_replace or []
        self.documents_to_delete = documents_to_delete or []
        self.delete_collection = delete_collection

    def execute(self, context: Context):
        hook = ArangoDBHook(arangodb_conn_id=self.arangodb_conn_id)

        if not any(
            [
                self.documents_to_insert,
                self.documents_to_update,
                self.documents_to_replace,
                self.documents_to_delete,
                self.delete_collection,
            ]
        ):
            raise AirflowException("At least one operation must be specified.")

        if self.documents_to_insert:
            self.log.info(
                "Inserting %d documents into collection '%s'.",
                len(self.documents_to_insert),
                self.collection_name,
            )
            hook.insert_documents(self.collection_name, self.documents_to_insert)

        if self.documents_to_update:
            self.log.info(
                "Updating %d documents in collection '%s'.",
                len(self.documents_to_update),
                self.collection_name,
            )
            hook.update_documents(self.collection_name, self.documents_to_update)

        if self.documents_to_replace:
            self.log.info(
                "Replacing %d documents in collection '%s'.",
                len(self.documents_to_replace),
                self.collection_name,
            )
            hook.replace_documents(self.collection_name, self.documents_to_replace)

        if self.documents_to_delete:
            self.log.info(
                "Deleting %d documents from collection '%s'.",
                len(self.documents_to_delete),
                self.collection_name,
            )
            hook.delete_documents(self.collection_name, self.documents_to_delete)

        if self.delete_collection:
            self.log.info("Deleting collection '%s'.", self.collection_name)
            hook.delete_collection(self.collection_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.arangodb.operators.arangodb import AQLOperator, ArangoDBCollectionOperator


class TestAQLOperator:
    @mock.patch("airflow.providers.arangodb.operators.arangodb.ArangoDBHook")
    def test_arangodb_operator_test(self, mock_hook):
        arangodb_query = "FOR doc IN students RETURN doc"
        op = AQLOperator(task_id="basic_aql_task", query=arangodb_query)
        op.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(arangodb_conn_id="arangodb_default")
        mock_hook.return_value.query.assert_called_once_with(arangodb_query)


class TestArangoDBCollectionOperator:
    @mock.patch("airflow.providers.arangodb.operators.arangodb.ArangoDBHook")
    def test_insert_documents(self, mock_hook):
        documents_to_insert = [{"_key": "lola", "first": "Lola", "last": "Martin"}]
        op = ArangoDBCollectionOperator(
            task_id="insert_task",
            collection_name="students",
            documents_to_insert=documents_to_insert,
            documents_to_update=None,
            documents_to_replace=None,
            documents_to_delete=None,
            delete_collection=False,
        )
        op.execute(mock.MagicMock())
        mock_hook.assert_called_once_with(arangodb_conn_id="arangodb_default")
        mock_hook.return_value.insert_documents.assert_called_once_with("students", documents_to_insert)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from airflow.models import Connection
from airflow.models.dag import DAG
from airflow.providers.arangodb.sensors.arangodb import AQLSensor
from airflow.utils import timezone

# The tests do not create dag runs, so db isolation tests are skipped

DEFAULT_DATE = timezone.datetime(2017, 1, 1)
arangodb_hook_mock = Mock(name="arangodb_hook_for_test", **{"query.return_value.count.return_value": 1})


class TestAQLSensor:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        dag = DAG("test_dag_id", schedule=None, default_args=args)
        self.dag = dag
        create_connection_without_db(
            Connection(
                conn_id="arangodb_default",
                conn_type="arangodb",
                host="http://127.0.0.1:8529",
                login="root",
                password="password",
                schema="_system",
            )
        )

    @patch(
        "airflow.providers.arangodb.sensors.arangodb.ArangoDBHook",
        autospec=True,
        return_value=arangodb_hook_mock,
    )
    def test_arangodb_document_created(self, arangodb_mock):
        query = "FOR doc IN students FILTER doc.name == 'judy' RETURN doc"

        arangodb_tag_sensor = AQLSensor(
            task_id="aql_search_document",
            query=query,
            timeout=60,
            poke_interval=10,
            dag=self.dag,
        )

        arangodb_tag_sensor.execute({})
        assert arangodb_hook_mock.query.return_value.count.called



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module allows connecting to a ArangoDB."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

from arango import AQLQueryExecuteError, ArangoClient as ArangoDBClient
from arango.cursor import Cursor
from arango.exceptions import (
    DocumentDeleteError,
    DocumentInsertError,
    DocumentReplaceError,
    DocumentUpdateError,
)

from airflow.exceptions import AirflowException

try:
    from airflow.sdk import BaseHook
except ImportError:
    from airflow.hooks.base import BaseHook  # type: ignore[attr-defined,no-redef]

if TYPE_CHECKING:
    from arango.database import StandardDatabase

    from airflow.models import Connection


class ArangoDBHook(BaseHook):
    """
    Interact with ArangoDB.

    Performs a connection to ArangoDB and retrieves client.

    :param arangodb_conn_id: Reference to :ref:`ArangoDB connection id <howto/connection:arangodb>`.
    """

    conn_name_attr = "arangodb_conn_id"
    default_conn_name = "arangodb_default"
    conn_type = "arangodb"
    hook_name = "ArangoDB"

    def __init__(self, arangodb_conn_id: str = default_conn_name, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.arangodb_conn_id = arangodb_conn_id

    @cached_property
    def client(self) -> ArangoDBClient:
        """Initiates a new ArangoDB connection (cached)."""
        return ArangoDBClient(hosts=self.hosts)

    @cached_property
    def db_conn(self) -> StandardDatabase:
        """Connect to an ArangoDB database and return the database API wrapper."""
        return self.client.db(name=self.database, username=self.username, password=self.password)

    @cached_property
    def _conn(self) -> Connection:
        return self.get_connection(self.arangodb_conn_id)  # type: ignore[return-value]

    @property
    def hosts(self) -> list[str]:
        if not self._conn.host:
            raise AirflowException(f"No ArangoDB Host(s) provided in connection: {self.arangodb_conn_id!r}.")
        return self._conn.host.split(",")

    @property
    def database(self) -> str:
        if not self._conn.schema:
            raise AirflowException(f"No ArangoDB Database provided in connection: {self.arangodb_conn_id!r}.")
        return self._conn.schema

    @property
    def username(self) -> str:
        if not self._conn.login:
            raise AirflowException(f"No ArangoDB Username provided in connection: {self.arangodb_conn_id!r}.")
        return self._conn.login

    @property
    def password(self) -> str:
        return self._conn.password or ""

    def get_conn(self) -> ArangoDBClient:
        """Initiate a new ArangoDB connection (cached)."""
        return self.client

    def query(self, query, **kwargs) -> Cursor:
        """
        Create an ArangoDB session and execute the AQL query in the session.

        :param query: AQL query
        """
        try:
            if self.db_conn:
                result = self.db_conn.aql.execute(query, **kwargs)
                if not isinstance(result, Cursor):
                    raise AirflowException("Failed to execute AQLQuery, expected result to be of type Cursor")
                return result
            raise AirflowException(
                f"Failed to execute AQLQuery, error connecting to database: {self.database}"
            )
        except AQLQueryExecuteError as error:
            raise AirflowException(f"Failed to execute AQLQuery, error: {error}")

    def create_collection(self, name):
        if not self.db_conn.has_collection(name):
            self.log.info("Collection '%s' does not exist. Creating a new collection.", name)
            self.db_conn.create_collection(name)
            return True
        self.log.info("Collection already exists: %s", name)
        return False

    def delete_collection(self, name):
        if self.db_conn.has_collection(name):
            self.db_conn.delete_collection(name)
            return True
        self.log.info("Collection does not exist: %s", name)
        return False

    def create_database(self, name):
        if not self.db_conn.has_database(name):
            self.db_conn.create_database(name)
            return True
        self.log.info("Database already exists: %s", name)
        return False

    def create_graph(self, name):
        if not self.db_conn.has_graph(name):
            self.db_conn.create_graph(name)
            return True
        self.log.info("Graph already exists: %s", name)
        return False

    def insert_documents(self, collection_name, documents):
        if not self.db_conn.has_collection(collection_name):
            self.create_collection(collection_name)

        try:
            collection = self.db_conn.collection(collection_name)
            collection.insert_many(documents, silent=True)
        except DocumentInsertError as e:
            self.log.error("Failed to insert documents: %s", str(e))
            raise

    def update_documents(self, collection_name, documents):
        if not self.db_conn.has_collection(collection_name):
            raise AirflowException(f"Collection does not exist: {collection_name}")

        try:
            collection = self.db_conn.collection(collection_name)
            collection.update_many(documents, silent=True)
        except DocumentUpdateError as e:
            self.log.error("Failed to update documents: %s", str(e))
            raise

    def replace_documents(self, collection_name, documents):
        if not self.db_conn.has_collection(collection_name):
            raise AirflowException(f"Collection does not exist: {collection_name}")

        try:
            collection = self.db_conn.collection(collection_name)
            collection.replace_many(documents, silent=True)
        except DocumentReplaceError as e:
            self.log.error("Failed to replace documents: %s", str(e))
            raise

    def delete_documents(self, collection_name, documents):
        if not self.db_conn.has_collection(collection_name):
            raise AirflowException(f"Collection does not exist: {collection_name}")

        try:
            collection = self.db_conn.collection(collection_name)
            collection.delete_many(documents, silent=True)
        except DocumentDeleteError as e:
            self.log.error("Failed to delete documents: %s", str(e))
            raise

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        return {
            "hidden_fields": ["port", "extra"],
            "relabeling": {
                "host": "ArangoDB Host URL or  comma separated list of URLs (coordinators in a cluster)",
                "schema": "ArangoDB Database",
                "login": "ArangoDB Username",
                "password": "ArangoDB Password",
            },
            "placeholders": {
                "host": 'eg."http://127.0.0.1:8529" or "http://127.0.0.1:8529,http://127.0.0.1:8530"'
                " (coordinators in a cluster)",
                "schema": "_system",
                "login": "root",
                "password": "password",
            },
        }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.arangodb.hooks.arangodb import ArangoDBHook
from airflow.providers.arangodb.version_compat import BaseSensorOperator

if TYPE_CHECKING:
    try:
        from airflow.sdk.definitions.context import Context
    except ImportError:
        # TODO: Remove once provider drops support for Airflow 2
        from airflow.utils.context import Context


class AQLSensor(BaseSensorOperator):
    """
    Checks for the existence of a document which matches the given query in ArangoDB.

    :param collection: Target DB collection.
    :param query: The query to poke, or you can provide .sql file having the query
    :param arangodb_conn_id: The :ref:`ArangoDB connection id <howto/connection:arangodb>` to use
        when connecting to ArangoDB.
    :param arangodb_db: Target ArangoDB name.
    """

    template_fields: Sequence[str] = ("query",)
    template_ext: Sequence[str] = (".sql",)
    template_fields_renderers = {"query": "sql"}

    def __init__(self, *, query: str, arangodb_conn_id: str = "arangodb_default", **kwargs) -> None:
        super().__init__(**kwargs)
        self.arangodb_conn_id = arangodb_conn_id
        self.query = query

    def poke(self, context: Context) -> bool:
        self.log.info("Sensor running the following query: %s", self.query)
        hook = ArangoDBHook(self.arangodb_conn_id)
        records = hook.query(self.query, count=True).count()
        self.log.info("Total records found: %d", records)
        return records != 0



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import Mock, patch

import pytest
from arango.cursor import Cursor

from airflow.models import Connection
from airflow.providers.arangodb.hooks.arangodb import ArangoDBHook

arangodb_client_mock = Mock(name="arangodb_client_for_test")


class TestArangoDBHook:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id="arangodb_default",
                conn_type="arangodb",
                host="http://127.0.0.1:8529",
                login="root",
                password="password",
                schema="_system",
            )
        )

    @patch(
        "airflow.providers.arangodb.hooks.arangodb.ArangoDBClient",
        autospec=True,
        return_value=arangodb_client_mock,
    )
    def test_get_conn(self, arango_mock):
        arangodb_hook = ArangoDBHook()

        assert arangodb_hook.hosts == ["http://127.0.0.1:8529"]
        assert arangodb_hook.username == "root"
        assert arangodb_hook.password == "password"
        assert arangodb_hook.database == "_system"
        assert arangodb_hook.client is not None
        assert arango_mock.called
        assert isinstance(arangodb_hook.client, Mock)

    @patch(
        "airflow.providers.arangodb.hooks.arangodb.ArangoDBClient",
        autospec=True,
        return_value=arangodb_client_mock,
    )
    def test_query(self, arango_mock):
        arangodb_hook = ArangoDBHook()
        with patch.object(arangodb_hook, "db_conn"):
            mock_cursor = Mock(spec=Cursor)
            arangodb_hook.db_conn.aql.execute.return_value = mock_cursor

            arangodb_query = "FOR doc IN students RETURN doc"
            arangodb_hook.query(arangodb_query)

            assert arango_mock.called
            assert isinstance(arangodb_hook.client, Mock)
            assert arango_mock.return_value.db.called

    @patch(
        "airflow.providers.arangodb.hooks.arangodb.ArangoDBClient",
        autospec=True,
        return_value=arangodb_client_mock,
    )
    def test_create_database(self, arango_mock):
        arangodb_hook = ArangoDBHook()
        arangodb_hook.create_database(name="_system")

        arango_mock.return_value.db.return_value.has_database.return_value = False

        assert arango_mock.called
        assert isinstance(arangodb_hook.client, Mock)
        assert arango_mock.return_value.db.called
        assert arango_mock.return_value.db.return_value.has_database.called

    @patch(
        "airflow.providers.arangodb.hooks.arangodb.ArangoDBClient",
        autospec=True,
        return_value=arangodb_client_mock,
    )
    def test_create_collection(self, arango_mock):
        arangodb_hook = ArangoDBHook()
        arangodb_hook.create_collection(name="student")

        arango_mock.return_value.db.return_value.has_collection.return_value = False

        assert arango_mock.called
        assert isinstance(arangodb_hook.client, Mock)
        assert arango_mock.return_value.db.called
        assert arango_mock.return_value.db.return_value.has_collection.called

    @patch(
        "airflow.providers.arangodb.hooks.arangodb.ArangoDBClient",
        autospec=True,
        return_value=arangodb_client_mock,
    )
    def test_create_graph(self, arango_mock):
        arangodb_hook = ArangoDBHook()
        arangodb_hook.create_graph(name="student_network")

        arango_mock.return_value.db.return_value.has_graph.return_value = False

        assert arango_mock.called
        assert isinstance(arangodb_hook.client, Mock)
        assert arango_mock.return_value.db.called
        assert arango_mock.return_value.db.return_value.has_graph.called
